/*
 * mpeg2decoder.c: A plugin for the Video Disk Recorder
 *
 * See the README file for copyright information and how to reach the author.
 *
 * $Id: mpeg2decoder.c,v 1.4 2004/07/10 15:06:47 stefan Exp $
 */

#include <vdr/plugin.h>

#include "mpeg2decoder.h"
#include "audio.h"
#include "utils.h"
#include "setup-softdevice.h"
#define UNSYNCED 0
#define PAYLOAD 100
#define PAYLOADDATA 200
#define HEADER 300
#define OPTHEADER 400
#define STREAM 500




#define GET_MPEG2_PTS(x)   ( ((uint64_t)x[4]&0xFE) >>1     | \
     		    	     ((uint64_t)x[3])      <<7     | \
                             ((uint64_t)x[2]&0xFE) <<14    | \
                             ((uint64_t)x[1])      <<22    | \
                             ((uint64_t)x[0]&0x0E) <<29 )


// --- cStreamDecoder ---------------------------------------------------------

cStreamDecoder::cStreamDecoder(unsigned int StreamID)
{
    streamID = StreamID;
    frame=0;
}



cStreamDecoder::~cStreamDecoder() {
}

void cStreamDecoder::SyncPTS(uint64_t spts) {
    int diff= (int)(pts-spts);
    if (abs(diff) > 100) {
	pts=spts;
    }	
}

int cStreamDecoder::ParseStream(uchar *Data, int Length) {
  return ParseStreamIntern(Data,Length,streamID,streamID);
}

int cStreamDecoder::ParseStreamIntern(uchar *Data, int Length,
                                      unsigned int lowID,
                                      unsigned int highID) {
	int len, streamlen;
	uint8_t *inbuf_ptr;
        inbuf_ptr = (uint8_t *)Data;
	int size=Length;
        while (size > 0) {
	    len=1;
	    switch (state) {
		// here we found a sync header
		case UNSYNCED:
		    syncword = (syncword << 8) | *inbuf_ptr;
	    	    if ( syncword >= lowID && syncword <= highID) {
			state=HEADER;
		    } 
		break;    
		case HEADER:		//Payload length(hi)
		    payload = *inbuf_ptr << 8;
		    state++;
		break;
		case HEADER+1:		//Payload length(lo)
		    payload += *inbuf_ptr;
		    state++;
		break;

		case HEADER+2:		//wei ich nicht (immer 0x80 oder 0x85)
		    state++;  
		    payload--;  
		break;

		case HEADER+3:		// ??
		    state++;  
		    payload--;  
		break;
		
		case HEADER+4:		// PTS-Header length
		    headerLength=*inbuf_ptr;
		    state=STREAM;
		    hdr_ptr=0;
		    validPTS=false;

		    if (headerLength) { // we found optional headers (=PTS)
			validPTS=true;
			state=OPTHEADER;  
			hdr_ptr=header;
		    }
		    payload--;  
		break;
		
		case OPTHEADER:		// save PTS
		    payload--;
		    *hdr_ptr=*inbuf_ptr;
		    hdr_ptr++;
		    if (!--headerLength) state=STREAM;  
		break;
		
		case STREAM:
		    streamlen=min(payload,size); // Max data that is avaiable
		    len=DecodeData(inbuf_ptr,streamlen);
		    payload-=len;
		    if (payload<=0) {
			state=UNSYNCED;	
		    }
		break;    
		default:
		    state=UNSYNCED;
	    }
            size -= len;
            inbuf_ptr += len;
        }
	return Length;
}

// --- AUDIO ------------------------------------------------------------------
cAudioStreamDecoder::cAudioStreamDecoder(unsigned int StreamID,
                                         cAudioOut *AudioOut,
                                         uint64_t *commonPTS)
                                          : cStreamDecoder(StreamID)
{
    audioOut=AudioOut;
    cPTS=commonPTS;
    codec = avcodec_find_decoder(CODEC_ID_MP2);
    if (!codec) {
        printf("[mpegdecoder] Fatal error! Audio codec not found\n");
        exit(1);                                                                                          
    }
    context=avcodec_alloc_context();
    audiosamples=(uint8_t *)malloc(AVCODEC_MAX_AUDIO_FRAME_SIZE);
    if (avcodec_open(context, codec) < 0) {
        printf("[mpegdecoder] Fatal error! Could not open audio codec\n");
	exit(1);                                                                                          
    }
}    

void cAudioStreamDecoder::Write(uchar *Data, int Length) 
{

    int size = Length;
    while (size > 0) {
	int len = ParseStream(Data,size);
	Data += len;
	size -= len;
    }
} 

int cAudioStreamDecoder::DecodeData(uchar *Data, int Length) {
    int len;
    int audio_size;
    len=avcodec_decode_audio(context, (short *)audiosamples, &audio_size, Data, Length);

    if (audio_size > 0) {
	//fwrite(audiosamples,1,audio_size,fo);
  audioOut->SetParams(context->channels,context->sample_rate);
	audioOut->Write(audiosamples,audio_size);
	int delay = audioOut->GetDelay();
	if (delay < 20) { // if we have less than 20 ms in buffer we double frames
    	    audioOut->Write(audiosamples,audio_size);
	}
	pts += (audio_size/(48*4)); // PTS weiterzhlen, egal ob Samples gespielt oder nicht

//	printf("Audiodelay: %d \n",delay);
	*cPTS = pts - delay; // Das ist die Master-PTS die wird an den video Teil bergeben, 
	    // damit Video syncronisieren kann
	if (validPTS) {
    	    SyncPTS(GET_MPEG2_PTS(header)/90); // milisekunden
	}
    }
    return len;
}


cAudioStreamDecoder::~cAudioStreamDecoder()
{
    avcodec_close(context);
    free(context);
    free(audiosamples);
}

// --- VIDEO ------------------------------------------------------------------

cVideoStreamDecoder::cVideoStreamDecoder(unsigned int StreamID,
                                         cVideoOut *VideoOut,
                                         uint64_t *commonPTS)
                                          : cStreamDecoder(StreamID)
{
    codec = avcodec_find_decoder(CODEC_ID_MPEG2VIDEO);
    avgOffset=0;
    if (!codec) {
        printf("[mpegdecoder] Fatal error! Video codec not found\n");
        exit(1);
    }
    videoOut = VideoOut;
    cPTS=commonPTS;
    //syncdevice = SyncDevice;
    context=avcodec_alloc_context();
    picture=avcodec_alloc_frame();
    if(codec->capabilities&CODEC_CAP_TRUNCATED)
	context->flags|= CODEC_FLAG_TRUNCATED; /* we dont send complete frames */
    if (avcodec_open(context, codec) < 0) {
        printf("[mpegdecoder] Fatal error! Could not open video codec\n");
	exit(1);
    }
#if 0
    {
        int fmt, w, h;
        unsigned char *dest;

      if (VideoOut->GetInfo(&fmt,&dest,&w,&h)) {
        avpicture_fill(picture,dest,fmt,w,h);
      }
    }
#endif    
    ringBuffer=new cRingBufferLinear(1024*1024,1024,true);
    Start(); // starte thread
}


void cVideoStreamDecoder::Write(uchar *Data, int Length)
{
  int p;
  p = ringBuffer->Put(Data, Length);
#define BLOCK_BUFFER
#ifdef BLOCK_BUFFER
  while (p != Length) {
    //printf("ERROR: ring buffer overflow (%d bytes dropped)\n", Length - p);
    //printf("[mpegdecoder] Videobuffer is full. This should not happen! Audio is playing too slow\n");
    Length -= p;
    Data += p;
    p = ringBuffer->Put(Data, Length);
    usleep(1);
  }
#endif
}


void cVideoStreamDecoder::Action()
{
//    printf("Neuer Thread geatartet: pid:%d\n",getpid());
    running=true;
    active=true;
    while(active) {
      int size=10240;
      mutex.Lock();
      uchar *u =ringBuffer->Get(size);
      if (u) {
        size=ParseStream(u,size);
        ringBuffer->Del(size);
        mutex.Unlock();
      } else {
        mutex.Unlock();
        usleep(1000);
      }
    }
    running=false;
//    printf("Thread beendet\n");
}


void cVideoStreamDecoder::resetCodec(void) {
    printf("[mpegdecoder] resetting codec\n");
    avcodec_close(context);
    if(codec->capabilities&CODEC_CAP_TRUNCATED)
	context->flags|= CODEC_FLAG_TRUNCATED; /* we dont send complete frames */
    if (avcodec_open(context, codec) < 0) {
        printf("[mpegdecoder] Fatal error! Could not open video codec\n");
	exit(1);
    }
}

int cVideoStreamDecoder::StillPicture(uchar *Data, int Length) {
  mutex.Lock();
  ringBuffer->Clear();
  avcodec_flush_buffers(context);
  mutex.Unlock();
  for (int i = 0;i < 4;++i) {
    ParseStreamIntern(Data,Length,0x000001E0,0x000001EF);
  }
  return Length;
}

int cVideoStreamDecoder::DecodeData(uchar *Data, int Length) {
    int len;
    int got_picture;

    len = avcodec_decode_video(context, picture, &got_picture,Data, Length);
    if (len < 0) {
    	    printf("[mpegdecoder] Error while decoding frame %d\n", frame);
	    resetCodec();
    	    return Length;
    }
    if (got_picture) {
      if (setupStore.deintMethod == 1) {
        // deinterlace with internal libavcodec deinterlacer
        avpic.data[0]     = picture->data[0];
        avpic.data[1]     = picture->data[1];
        avpic.data[2]     = picture->data[2];
        avpic.linesize[0] = picture->linesize[0];
        avpic.linesize[1] = picture->linesize[1];
        avpic.linesize[2] = picture->linesize[2];

        if (avpicture_deinterlace(&avpic, &avpic, context->pix_fmt,
                                  context->width, context->height) < 0) {
          fprintf(stderr,
                  "[mpegdecoder] error, libavcodec deinterlacer failure\n");
        }
      }

      if (validPTS /*&& context->coded_frame->pict_type == FF_I_TYPE*/) {
        pts=(GET_MPEG2_PTS(header)/90);
      }

// this few lines does the whole syncing
	int offset = *cPTS - pts;
	if (offset > 1000) offset = 1000;
	if (offset < -1000) offset = -1000;

	avgOffset = (int)( ( (24*avgOffset) + offset ) / 25);
	if (avgOffset > 1000) avgOffset = offset;
	if (avgOffset < -1000) avgOffset = offset;

	int frametime = 40; //ms
	if (avgOffset > 20 && offset > 0) {
	    frametime -= (avgOffset / 10);
	}
	if (avgOffset < -20 && offset < 0) {
	    frametime -= (avgOffset / 10);
	}
#if 0
	if (!(frame % 10) || context->hurry_up) {
	    printf ("[mpegdecoder] Frame# %-5d A-V(ms) %-5d(Avg:%-4d) Frame Time:%-3d ms Hurry: %d\n",frame,(int)(*cPTS-pts),avgOffset, frametime,context->hurry_up);
	}
#endif
	//frametime=40;
	pts += 40;
    	vpts += frametime; // 40 ms for PAL   // FIXME FOR NTSC

	int delay;
        delay = vpts - getTimeMilis();
	if (delay < -1000) {
	    vpts = getTimeMilis();
	    context->hurry_up++;
	} else {
	    if (delay > 1000) {
		vpts = getTimeMilis();
	    } else {
		while (delay > 0) {
		    usleep(1000);
		    delay = vpts - getTimeMilis();
		}
	    }
	    if (context->hurry_up) context->hurry_up--;
	    if (context->hurry_up < 3) {
        videoOut->CheckAspectDimensions(picture,context);
        videoOut->YUV(picture->data[0], picture->data[1],picture->data[2],
                      context->width,context->height,
                      picture->linesize[0],picture->linesize[1]);
	    }
	}
	frame++;
    }
    return len;
}

cVideoStreamDecoder::~cVideoStreamDecoder()
{
    active = false;
    Cancel(3);
    delete(ringBuffer);
    avcodec_close(context);
    free(context);
    free(picture);
}

//----------------------------   create a new MPEG Decoder
cMpeg2Decoder::cMpeg2Decoder(cAudioOut *AudioOut, cVideoOut *VideoOut)

{
    avcodec_init();
    avcodec_register_all();
    state=UNSYNCED;
    audioOut=AudioOut;
    videoOut=VideoOut;
    aout=vout=0;
    running=false;
    decoding=false;
}

cMpeg2Decoder::~cMpeg2Decoder() {
    if (aout) delete(aout);
    aout=0;
    if (vout) delete(vout);
    vout=0;
}

void cMpeg2Decoder::Start(void)
{
    // ich wei nicht, ob man's so kompliziert machen soll, aber jetzt hab ich's
    // schon so gemacht, das 2 Threads gestartet werden.
    // Audio is the master, Video syncs on Audio
    aout = new cAudioStreamDecoder( 0x000001C0, audioOut, &commonPTS);
    vout = new cVideoStreamDecoder( 0x000001E0, videoOut, &commonPTS);
    running=true;
}

void cMpeg2Decoder::Stop(void)
{

    if (running) {
	running=false;
	while (decoding) {
	    usleep(1);
	}
	if (vout) {
//	    vout->Stop();
	    delete(vout);
	}

	if (aout) {
//	    aout->Stop();
	    delete(aout);
	}
	aout=vout=0;
    }
}

int cMpeg2Decoder::StillPicture(uchar *Data, int Length) {
//  vout->StillPicture(Data,Length);
//  vout->StillPicture(Data,Length);
  return vout->StillPicture(Data,Length);
//  return Decode(Data,Length);
}

int cMpeg2Decoder::Decode(const uchar *Data, int Length) {

    int len, streamlen;
    uint8_t *inbuf_ptr;
    inbuf_ptr = (uint8_t *)Data;
    int size=Length;
    decoding=true;
    if (!running) {
        decoding=false;
	return Length;
    }
    while (size > 0) {
	len=1;
	switch (state) {
	    case UNSYNCED:
		syncword = (syncword << 8) | *inbuf_ptr;
	    	if ( (syncword >= 0x000001E0) && (syncword <= 0x000001EF) ) {
		    state=PAYLOAD;
		    streamtype=0xE0;
		} else
		if ( (syncword >= 0x000001C0) && (syncword <= 0x000001CF) ) {
		    state=PAYLOAD;
		    streamtype=0xC0;
		}
		break;
	    case PAYLOAD:		//Payload length(hi)
		payload = *inbuf_ptr << 8;
		state++;
	    break;
	    case PAYLOAD+1:		//Payload length(lo)
		payload += *inbuf_ptr;
		header[0]=0;
		header[1]=0;
		header[2]=1;
		header[3]=streamtype;
		header[4]=payload >> 8;
		header[5]=payload & 0xFF;
		if (streamtype == 0xE0) {
		    vout->Write(header,6);
		}
		if (streamtype == 0xC0) {
		    aout->Write(header,6);
		}
		state=PAYLOADDATA;
	    break;
	    case PAYLOADDATA:

		streamlen =min(payload,size);
		if (streamtype == 0xE0) {
		    vout->Write(inbuf_ptr,streamlen);
		}
		if (streamtype == 0xC0) {
		    aout->Write(inbuf_ptr,streamlen);
		}
		payload-=streamlen;
		len = streamlen;
		if (payload <= 0) state=UNSYNCED;
	    break;
	}
    	size -= len;
	inbuf_ptr += len;
    }
    decoding=false;
    return Length;
}


