/*
 * video.c: A plugin for the Video Disk Recorder
 *
 * See the README file for copyright information and how to reach the author.
 *
 * $Id: video-dfb.c,v 1.5 2004/07/10 20:24:03 stefan Exp $
 */

#include <sys/mman.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <vdr/plugin.h>
#include "video-dfb.h"
#include "utils.h"

#define HAVE_SetSourceLocation 0

//#define COLORKEY 17,8,79
#define COLORKEY 0,0,0

/* macro for a safe call to DirectFB functions */
#define DFBCHECK(x...) \
     {                                                               \
           DFBResult  err = x;                                       \
           if (err != DFB_OK) {                                      \
              fprintf( stderr, "%s <%d>:\n\t", __FILE__, __LINE__ ); \
              DirectFBErrorFatal( #x, err );                         \
           }                                                         \
     }

//static cDFBRemote           *dfbRemote = NULL;

// --- cDFBRemote ---------------------------------------------------
/* ---------------------------------------------------------------------------
 */
cDFBRemote::cDFBRemote(const char *Name, cDFBVideoOut *vout) : cRemote(Name)
{
  video_out = vout;
  /* create an event buffer with all devices attached */
  events = vout->dfb->CreateInputEventBuffer( DICAPS_ALL );
}

/* ---------------------------------------------------------------------------
 */
cDFBRemote::~cDFBRemote()
{
  active = false;
  Cancel(2);
}

/* ---------------------------------------------------------------------------
 */
void cDFBRemote::PutKey(DFBInputDeviceKeySymbol key)
{
  Put ((int) key);
}

/* ---------------------------------------------------------------------------
 */
void cDFBRemote::Action(void)
{
    DFBInputEvent event;

  dsyslog("DFB remote control thread started (pid=%d)", getpid());
  active = true;
  while (active)
  {
    //usleep (25000);
    events->WaitForEvent();

    /* handle all events, exit if HandleEvent() returns true */
    while (events->GetEvent( DFB_EVENT(&event) ))
      video_out->ProcessEvents (event);

  }
  dsyslog("DFB remote control thread ended (pid=%d)", getpid());
}

/* ---------------------------------------------------------------------------
 */
void cDFBRemote::DFBRemoteStart(void)
{
  Start();
}

// --- cDFBVideoOut -------------------------------------------------
IDirectFB *DFBWrapper;

/* ---------------------------------------------------------------------------
 */
static DFBEnumerationResult EnumCallBack( unsigned int id, DFBDisplayLayerDescription desc, void *data )
{
    IDirectFBDisplayLayer **layer = (IDirectFBDisplayLayer **)data;

  fprintf(stderr,"Layer %d %s",id, desc.name);

  fprintf(stderr,"  Type: ");
  if (desc.type & DLTF_BACKGROUND) fprintf(stderr,"background " );
  if (desc.type & DLTF_GRAPHICS) fprintf(stderr,"graphics " );
  if (desc.type & DLTF_STILL_PICTURE) fprintf(stderr,"picture " );
  if (desc.type & DLTF_VIDEO) fprintf(stderr,"video " );
  fprintf(stderr,"\n" );

  fprintf(stderr,"  Caps: " );
  if (desc.caps & DLCAPS_ALPHACHANNEL) fprintf(stderr,"alphachannel " );
  if (desc.caps & DLCAPS_BRIGHTNESS) fprintf(stderr,"brightness " );
  if (desc.caps & DLCAPS_CONTRAST) fprintf(stderr,"contrast " );
  if (desc.caps & DLCAPS_DEINTERLACING) fprintf(stderr,"deinterlacing " );
  if (desc.caps & DLCAPS_DST_COLORKEY) fprintf(stderr,"dst_colorkey " );
  if (desc.caps & DLCAPS_FLICKER_FILTERING) fprintf(stderr,"flicker_filtering " );
  if (desc.caps & DLCAPS_HUE) fprintf(stderr,"hue " );
  if (desc.caps & DLCAPS_LEVELS) fprintf(stderr,"levels " );
  if (desc.caps & DLCAPS_OPACITY) fprintf(stderr,"opacity " );
  if (desc.caps & DLCAPS_SATURATION) fprintf(stderr,"saturation " );
  if (desc.caps & DLCAPS_SCREEN_LOCATION) fprintf(stderr,"screen_location " );
  if (desc.caps & DLCAPS_SRC_COLORKEY) fprintf(stderr,"src_colorkey " );
  if (desc.caps & DLCAPS_SURFACE) fprintf(stderr,"surface " );
  fprintf(stderr,"\n");

  /* We take the first layer not being the primary */
  if (id != DLID_PRIMARY && desc.caps & DLCAPS_SURFACE && !*layer) {
    fprintf(stderr,"  This is our videoLayer\n");
    try {
      *layer = DFBWrapper->GetDisplayLayer(id);
    } catch (DFBException *ex) {
      fprintf(stderr,"Caught: %s", ex);
      return DFENUM_CANCEL;
    }
  }

  return DFENUM_OK;
}

/* ---------------------------------------------------------------------------
 */
// List Video Modes
static DFBEnumerationResult EnumVideoModeCallback(int x, int y, int bpp, void *data)
{
  fprintf(stderr,"%dx%d@%d ",x,y,bpp);
  return DFENUM_OK;
}

/* ---------------------------------------------------------------------------
 */
cDFBVideoOut::cDFBVideoOut()
{
    DFBSurfaceDescription dsc;
    DFBDisplayLayerDescription   desc;
    DFBCardCapabilities caps;

  fprintf(stderr,"[dfb] init\n");
  DirectFB::Init();
  dfb = DirectFB::Create();
  dfb->SetCooperativeLevel(DFSCL_EXCLUSIVE);

  dfb->GetCardCapabilities(&caps);

  fprintf(stderr,"[dfb] RAM: %d bytes\n",caps.video_memory);

  fprintf(stderr,"[dfb] Accellerated Functions: ");
  if (caps.acceleration_mask & DFXL_NONE) fprintf(stderr,"none ");
  if (caps.acceleration_mask & DFXL_FILLRECTANGLE) fprintf(stderr,"FillRectange ");
  if (caps.acceleration_mask & DFXL_DRAWRECTANGLE) fprintf(stderr,"DrawRectange ");
  if (caps.acceleration_mask & DFXL_DRAWLINE) fprintf(stderr,"DrawLine ");
  if (caps.acceleration_mask & DFXL_FILLTRIANGLE) fprintf(stderr,"FillTriangle ");
  if (caps.acceleration_mask & DFXL_BLIT) fprintf(stderr,"Blit ");
  if (caps.acceleration_mask & DFXL_STRETCHBLIT) fprintf(stderr,"StretchBlit ");
  if (caps.acceleration_mask & DFXL_ALL) fprintf(stderr,"All ");
  fprintf(stderr,"\n");

  fprintf(stderr,"[dfb] Drawing Flags: ");
  if (caps.drawing_flags == DSDRAW_NOFX ) fprintf(stderr,"none ");
  if (caps.drawing_flags & DSDRAW_BLEND ) fprintf(stderr,"Blend ");
  if (caps.drawing_flags & DSDRAW_DST_COLORKEY ) fprintf(stderr,"Colorkeying ");
  if (caps.drawing_flags & DSDRAW_SRC_PREMULTIPLY ) fprintf(stderr,"Src.premultiply ");
  if (caps.drawing_flags & DSDRAW_DST_PREMULTIPLY ) fprintf(stderr,"Dst.premultiply ");
  if (caps.drawing_flags & DSDRAW_DEMULTIPLY ) fprintf(stderr,"Demultiply ");
  if (caps.drawing_flags & DSDRAW_XOR ) fprintf(stderr,"Xor ");
  fprintf(stderr,"\n");

  fprintf(stderr,"[dfb] Surface Blitting Flags: ");
  if (caps.blitting_flags == DSBLIT_NOFX ) fprintf(stderr,"none ");
  if (caps.blitting_flags & DSBLIT_BLEND_ALPHACHANNEL ) fprintf(stderr,"BlendAlpha ");
  if (caps.blitting_flags & DSBLIT_BLEND_COLORALPHA ) fprintf(stderr,"BlendColorAlpha ");
  if (caps.blitting_flags & DSBLIT_COLORIZE ) fprintf(stderr,"Colorize ");
  if (caps.blitting_flags & DSBLIT_SRC_COLORKEY ) fprintf(stderr,"SrcColorkey ");
  if (caps.blitting_flags & DSBLIT_DST_COLORKEY ) fprintf(stderr,"DstColorkey ");
  if (caps.blitting_flags & DSBLIT_SRC_PREMULTIPLY ) fprintf(stderr,"SrcPremultiply ");
  if (caps.blitting_flags & DSBLIT_DST_PREMULTIPLY ) fprintf(stderr,"DstPremultiply ");
  if (caps.blitting_flags & DSBLIT_DEMULTIPLY ) fprintf(stderr,"Demultiply ");
  if (caps.blitting_flags & DSBLIT_DEINTERLACE ) fprintf(stderr,"Deinterlace ");
  fprintf(stderr,"\n");

  fprintf(stderr,"[dfb] Supported video Modes are: ");
  dfb->EnumVideoModes(EnumVideoModeCallback, NULL);
  fprintf(stderr,"\n");

  Xres = 768;
  Yres = 576;
  Bpp  = 32;
#if 1
  Xres = 1280;
  Yres = 1024;
  Bpp  = 32;
#endif
  try {
    fprintf(stderr,"[dfb] Setting  Videomode to %dx%d-%d\n",Xres,Yres,Bpp);
    dfb->SetVideoMode(Xres,Yres,Bpp); // 768 x 576 @ 32bpp is the best choice for TV-Out
  } catch (DFBException *ex) {
    try {
      fprintf(stderr,"[dfb] This didn't work. Trying Videomode 1024x768-32\n");
      Xres = 1024;
      Yres = 768;
      Bpp  = 32;
      dfb->SetVideoMode(Xres,Yres,Bpp); // evtl hier andere Modes probieren
    } catch (DFBException *ex) {
      try {
        fprintf(stderr,"[dfb] This didn't work. Trying Videomode 1280x1024-32\n");
        Xres = 1280;
        Yres = 1024;
        Bpp  = 32;
        dfb->SetVideoMode(Xres,Yres,Bpp); // evtl hier andere Modes probieren
      } catch (DFBException *ex) {
        fprintf(stderr,"[dfb] This Vidmode doen't work, too. Exiting...\n");
        exit(1);
      }
    }
  }

  osdLayer=dfb->GetDisplayLayer(DLID_PRIMARY); // erster Layer. Fuer OSD

  dsc.flags = DSDESC_CAPS;
  DFB_ADD_SURFACE_CAPS(dsc.caps, DSCAPS_PRIMARY);
  DFB_ADD_SURFACE_CAPS(dsc.caps, DSCAPS_FLIPPING);

  osdSurface=dfb->CreateSurface(dsc);
  osdSurface->Clear(COLORKEY,0); //clear and
  osdSurface->Flip(); // Flip the field
  osdSurface->Clear(COLORKEY,0); //clear and
  osdSurface->Flip(); // Flip the field

  DFBWrapper=dfb;
  fprintf(stderr,"[dfb] Enumeratig display Layers\n");

  videoLayer = NULL;
  dfb->EnumDisplayLayers(EnumCallBack, &videoLayer);

  if (videoLayer) {
    videoSurface=videoLayer->GetSurface();
    videoSurface->Clear(COLORKEY,0); //clear and
    videoSurface->Flip(); // Flip the field
    videoSurface->Clear(COLORKEY,0); //clear and
    videoSurface->Flip(); // Flip the field
  } else {
    fprintf(stderr,"could not find suitable videolayer\n");
    exit(1);
  }

  if (!osdLayer) {
    fprintf(stderr,"[dfb] no OSD layer exiting\n");
    exit(1);
  }
  desc = osdLayer->GetDescription();
  fprintf(stderr,"Using this layer for OSD: %s\n", desc.name);
  fprintf(stderr,"Size: %d x %d \n",osdSurface->GetWidth(), osdSurface->GetHeight());

  DFBSurfaceCapabilities scaps;

  scaps = osdSurface->GetCapabilities();
  fprintf(stderr,"Caps:\n");
  if (scaps & DSCAPS_NONE) fprintf(stderr," - none\n");
  if (scaps & DSCAPS_PRIMARY) fprintf(stderr," - primary\n");
  if (scaps & DSCAPS_SYSTEMONLY) fprintf(stderr," - systemonly\n");
  if (scaps & DSCAPS_VIDEOONLY) fprintf(stderr," - videoonly\n");
  if (scaps & DSCAPS_FLIPPING) fprintf(stderr," - flipping\n");
  if (scaps & DSCAPS_SUBSURFACE) fprintf(stderr," - subsurface\n");
  if (scaps & DSCAPS_INTERLACED) fprintf(stderr," - interlaced\n");
  if (scaps & DSCAPS_SEPARATED) fprintf(stderr," - separated\n");
  if (scaps & DSCAPS_STATIC_ALLOC) fprintf(stderr," - static alloc\n");
  if (scaps & DSCAPS_TRIPLE) fprintf(stderr," - triple buffered\n");

  scaps = videoSurface->GetCapabilities();
  fprintf(stderr,"VideoCaps:\n");
  if (scaps & DSCAPS_NONE) fprintf(stderr," - none\n");
  if (scaps & DSCAPS_PRIMARY) fprintf(stderr," - primary\n");
  if (scaps & DSCAPS_SYSTEMONLY) fprintf(stderr," - systemonly\n");
  if (scaps & DSCAPS_VIDEOONLY) fprintf(stderr," - videoonly\n");
  if (scaps & DSCAPS_FLIPPING) fprintf(stderr," - flipping\n");
  if (scaps & DSCAPS_SUBSURFACE) fprintf(stderr," - subsurface\n");
  if (scaps & DSCAPS_INTERLACED) fprintf(stderr," - interlaced\n");
  if (scaps & DSCAPS_SEPARATED) fprintf(stderr," - separated\n");
  if (scaps & DSCAPS_STATIC_ALLOC) fprintf(stderr," - static alloc\n");
  if (scaps & DSCAPS_TRIPLE) fprintf(stderr," - triple buffered\n");

  // evtl wird die eine oder andere Funktion nicht von der Graka unterstuetzt
  fprintf(stderr,"[dfb] Configuring CooperativeLevel for Overlay\n");
  videoLayer->SetCooperativeLevel(DLSCL_ADMINISTRATIVE);

  fprintf(stderr,"[dfb] Configuring ColorKeying\n");
  videoLayer->SetDstColorKey(COLORKEY);

  fprintf(stderr,"[dfb] Configuring CooperativeLevel for OSD\n");
  osdLayer->SetCooperativeLevel(DLSCL_ADMINISTRATIVE);
  fprintf(stderr,"... configured\n");

  // Wenn wir hier sind. haben wir den grten Teil schon geschafft
  desc = osdLayer->GetDescription();
  fprintf(stderr,"[dfb] Using this layer for OSD: %s\n", desc.name);

  desc = videoLayer->GetDescription();
  fprintf(stderr,"[dfb] Using this layer for Video out: %s\n", desc.name);

  /* --------------------------------------------------------------------------
   * default settings: source, destination and logical widht/height
   * are set to our well known dimensions (except fwidth+fheight).
   */
  fwidth = 720;
  fheight = 576;

  lwidth = dwidth = swidth = Xres;
  lheight = dheight = sheight = Yres;
  sxoff = syoff = lxoff = lyoff = 0;

}

/* ---------------------------------------------------------------------------
 */
bool cDFBVideoOut::Initialize (void)
{
  /* -------------------------------------------------------------------------
   * force reconfiguring of video layer so cle266 color key works
   */
  aspect_changed = 1;
  SetParams ();
  aspect_changed = 0;
  /* -------------------------------------------------------------------------
   * get up our remote running
   */
  if (!dfbRemote)
  {
    dfbRemote = new cDFBRemote ("softdevice-dfb", this);
    dfbRemote->DFBRemoteStart();
  }
  return true;
}

/* ---------------------------------------------------------------------------
 */
void cDFBVideoOut::ProcessEvents (DFBInputEvent &event)
{
  switch (event.type) {
    case DIET_KEYRELEASE:
      if (dfbRemote)
        dfbRemote->PutKey(event.key_symbol);
      break;
    default:
      break;
  }
}

/* ---------------------------------------------------------------------------
 */
void cDFBVideoOut::SetParams()
{
    DFBDisplayLayerConfig       dlc;
    DFBDisplayLayerDescription  desc;

  if (videoLayer)
  {
    if (!videoSurface || aspect_changed)
    {
      fprintf(stderr,"[dfb] (re)configuring Videolayer to %d x %d \n", fwidth,fheight);
      dlc.flags   = (DFBDisplayLayerConfigFlags)(DLCONF_WIDTH | DLCONF_HEIGHT | DLCONF_PIXELFORMAT | DLCONF_OPTIONS);

#if HAVE_SetSourceLocation
      /* ----------------------------------------------------------------------
       * this should be the settings if SetSourceLocation is working for
       * all drivers.
       */
      dlc.width   = fwidth;
      dlc.height  = fheight;
      videoLayer->SetSourceLocation (sxoff, syoff, swidth, sheight);
#else
      /* ----------------------------------------------------------------------
       * for now we have to do another trick
       */
      dlc.width   = swidth;
      dlc.height  = sheight;
#endif
//      dlc.pixelformat = DSPF_YV12;
      dlc.pixelformat = DSPF_I420;

// DON'T use this format since it requires mmx planar -> packed conversion
//      dlc.pixelformat = DSPF_YUY2;

      pixelformat = dlc.pixelformat;

      desc = videoLayer->GetDescription();
//      fprintf (stderr, "layer caps = 0x%08x\n", desc.caps);
      if (desc.caps & DLCAPS_SCREEN_LOCATION)
      {
//        videoLayer->SetScreenLocation(0.0f, 0.0f, 0.9999f, 1.0f); // my Matrox doesn't wor when setting to 0.0 0.0 1.0 1.0
        videoLayer->SetScreenLocation((float) lxoff / (float) dwidth,
                                      (float) lyoff / (float) dheight,
                                      (float) lwidth / (float) dwidth,
                                      (float) lheight / (float) dheight);
      }
      else
      {
        fprintf(stderr,"Can't configure ScreenLocation. Hope it is Fullscreen\n");
      }

      /* ----------------------------------------------------------------------
       * set the default options to none
       */
      dlc.options = (DFBDisplayLayerOptions)( DLOP_NONE );

      if (desc.caps & DLCAPS_DEINTERLACING)
      {
//        fprintf (stderr, "deinterlacing\n");
        /* --------------------------------------------------------------------
         * if deinterlacing is supported we'll use that
         */
        dlc.options = (DFBDisplayLayerOptions)( DLOP_DEINTERLACING );
      }

      if (desc.caps & DLCAPS_ALPHACHANNEL)
      {
        /* --------------------------------------------------------------------
         * use alpha channel if it is supported
         */
        dlc.options = (DFBDisplayLayerOptions)((int)dlc.options|DLOP_ALPHACHANNEL);
      }
      else
      {
//        fprintf (stderr, "no alpha channel\n");
        if (desc.caps & DLCAPS_DST_COLORKEY)
        {
//          fprintf (stderr, "dest color key\n");
          /* ------------------------------------------------------------------
           * no alpha channel but destination color keying is supported
           */
          dlc.options = (DFBDisplayLayerOptions)((int)dlc.options|DLOP_DST_COLORKEY);
        }
      }

      /* ----------------------------------------------------------------------
       * OK, try to set the video layer configuration
       */
      try
      {
        videoLayer->SetConfiguration(dlc);
      }
      catch (DFBException *ex)
      {
        fprintf(stderr,"Caught: %s", ex);
        exit(1);
      }

      /* ----------------------------------------------------------------------
       * set colorkey now. some driver accepct that _after_ configuration
       * has been set !
       */
      videoLayer->SetDstColorKey(COLORKEY);

      videoSurface=videoLayer->GetSurface();
      fprintf(stderr,"[dfb] (re)configured 0x%08x\n",pixelformat);
    }
  }
  else
  {
    fprintf(stderr,"No Videolayer avaiable. Exiting...\n");
    exit(1);
  }
}

/* ---------------------------------------------------------------------------
 */
void cDFBVideoOut::Pause(void)
{
}

#if VDRVERSNUM >= 10307

/* ---------------------------------------------------------------------------
 */
void cDFBVideoOut::Refresh(cBitmap *Bitmap)
{
    int pitch;
    uint8_t *dst;

  osdSurface->Lock(DSLF_WRITE, (void **)&dst, &pitch) ;

  Draw(Bitmap,dst,pitch);

  osdSurface->Unlock();
  osdSurface->Flip();
}

#else

/* ---------------------------------------------------------------------------
 */
void cDFBVideoOut::Refresh()
{
    int pitch;
    uint8_t *dst;

  osdSurface->Lock(DSLF_WRITE, (void **)&dst, &pitch) ;
  for (int i = 0; i < MAXNUMWINDOWS; i++)
  {
    if (layer[i] && layer[i]->visible)
      layer[i]->Draw(dst, pitch, NULL);
  }
  osdSurface->Unlock();
  osdSurface->Flip();
}
#endif

/* ---------------------------------------------------------------------------
 */
void cDFBVideoOut::YUV(uint8_t *Py, uint8_t *Pu, uint8_t *Pv, int Width, int Height, int Ystride, int UVstride)
{
    uint8_t *dst;
    int pitch;
    int hi;

  SetParams();

  //fprintf(stderr,"[dfb] draw frame (%d x %d) Y: %d UV: %d\n", Width, Height, Ystride, UVstride);

  videoSurface->Lock(DSLF_WRITE, (void **)&dst, &pitch);
  if (pixelformat == DSPF_I420 || pixelformat == DSPF_YV12)
  {
#if HAVE_SetSourceLocation
    for(hi=0; hi < Height; hi++){
      memcpy(dst, Py, Width);
      Py  += Ystride;
      dst += pitch;
    }
    for(hi=0; hi < Height/2; hi++) {
      memcpy(dst, Pu, Width/2);
      Pu  += UVstride;
      dst += pitch / 2;
    }

    for(hi=0; hi < Height/2; hi++) {
      memcpy(dst, Pv, Width/2);
      Pv  += UVstride;
      dst += pitch / 2;
    }
#else
    for(hi=0; hi < sheight; hi++){
      memcpy(dst, Py+sxoff, swidth);
      Py  += Ystride;
      dst += pitch;
    }
    for(hi=0; hi < sheight/2; hi++) {
      memcpy(dst, Pu+sxoff/2, swidth/2);
      Pu  += UVstride;
      dst += pitch / 2;
    }

    for(hi=0; hi < sheight/2; hi++) {
      memcpy(dst, Pv+sxoff/2, swidth/2);
      Pv  += UVstride;
      dst += pitch / 2;
    }
#endif
  } else if (pixelformat == DSPF_YUY2) {

#ifndef USE_MMX
    /* reference implementation */
      int p = pitch - Width*2;
    for (int i=0; i<Height; i++) {
      for (int j =0; j < Width/2; j++) {
        *dst = *(Py + (j*2) + i * Ystride);
        dst +=1;
        *dst = *(Pu + j + (i/2) * UVstride);
        dst +=1;
        *dst = *(Py + (j*2)+1 + i * Ystride);
        dst +=1;
        *dst = *(Pv + j + (i/2) * UVstride);
        dst +=1;
      }
      dst +=p;
    }
# else
    /* deltas */
    for (int i=0; i<Height; i++) {
        uint8_t *pu, *pv, *py, *srfc;

      pu = Pu;
      pv = Pv;
      py = Py;
      srfc = dst;
      for (int j =0; j < Width/8; j++) {
        movd_m2r(*pu, mm1);       // mm1 = 00 00 00 00 U3 U2 U1 U0
        movd_m2r(*pv, mm2);       // mm2 = 00 00 00 00 V3 V2 V1 V0
        movq_m2r(*py, mm0);       // mm0 = Y7 Y6 Y5 Y4 Y3 Y2 Y1 Y0
        punpcklbw_r2r(mm2, mm1);  // mm1 = V3 U3 V2 U2 V1 U1 V0 U0
        movq_r2r(mm0,mm3);        // mm3 = Y7 Y6 Y5 Y4 Y3 Y2 Y1 Y0
        movq_r2r(mm1,mm4);        // mm4 = V3 U3 V2 U2 V1 U1 V0 U0
        punpcklbw_r2r(mm1, mm0);  // mm0 = V1 Y3 U1 Y2 V0 Y1 U0 Y0
        punpckhbw_r2r(mm4, mm3);  // mm3 = V3 Y7 U3 Y6 V2 Y5 U2 Y4

        movntq(mm0,*srfc);        // Am Meisten brauchen die Speicherzugriffe
        srfc+=8;
        py+=8;
        pu+=4;
        pv+=4;
        movntq(mm3,*srfc);      // wenn movntq nicht geht, dann movq verwenden
        srfc+=8;
      }
      Py += Ystride;;
      if (i % 2 == 1) {
        Pu += UVstride;
        Pv += UVstride;
      }
      dst += pitch;
    }
# endif
  }
  videoSurface->Unlock();

/*
  try {
    videoSurface->Flip();
  } catch (DFBException *ex){
    fprintf(stderr,"Flip failed\n");
  }
*/
}

/* ---------------------------------------------------------------------------
 */
void cDFBVideoOut::CloseOSD()
{
    osdSurface->Clear(COLORKEY,0); //clear and
    osdSurface->Flip(); // Flip the field
    osdSurface->Clear(COLORKEY,0); //clear and
    osdSurface->Flip(); // Flip the field
}

/* ---------------------------------------------------------------------------
 */
cDFBVideoOut::~cDFBVideoOut()
{
    fprintf(stderr,"Releasing DFB\n");
    if (dfb) dfb->Release();
}
